#ifndef __CONSUME_DATA__
#define __CONSUME_DATA__

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "read_sync.h"		// holds data structure to syncronize reader/consumer behaviour
#include "consumer_sync.h"
#ifdef WIN32
#include <winbase.h>
#endif
#ifdef LINUX
#include <string.h>
#endif

extern struct Fam *families;					// need to declare those guys "extern" in class header files
extern struct Read_sync *read_synchronizer;
extern struct Result *result;
extern float **d_matrix;										// For permutation only. d_matrix[num_pedigrees][SNPs]	

const int affection_field = 5;				// the field that stores the affection status
const int max_allele = 16;
const int max_geno = 136;
const int affected = 2;
const int unaffected = 1;
const int largest_pedigee = 200;
const int largest_nucfam = 20;

struct Temp_storage{						// This struct is pretty generic. Each thread stores its intermediate (for dis sibs)
	int **xst;								// Pedigree results here. In the End (of each pedigree) this is used to calculate: D=[1/(nt+ns)]*{sum_nt(xt) + sum_ns(xs)}
											// here we combine xt & xs into xst for efficiency reasons (see paper for notation)
	int **yst;								// ...as above for triads
	float **d;								// It is more efficient than competing for mutexes all the time. In the END all threads combine their partial result into one.
	float **dsq;							// we need to keep track of D & D^2
	int **used;								// keep track how many times an allele was used. Can't include zero useage alleles in final result.
	int *nt;
	int *ns;								// xt,xs,nt,ns are as in PDT-paper.
	int *ntf;
	int *nsf;								// same on per pedigree basis
	int *fams;								// family usage counter per locus
	int *useage;							// indicate if a family was used for the calculations
	int **used_sibs;						// keeps track
	// PDT-formula is:
	// D = [1/(nt+ns)]*{sum_nt(xt) + sum_ns(xs)} (calculated for EACH individual pedigree, over all possible triads & dis.sib-pairs)
	// nt = # of triads
	// ns = # of discordant siblings
	// xt = (# of allele_n transmitted) - (# of allele_n not transmitted)
	// xs = (# of allele_n in aff. sib) - (# of allele_n in unaff. sib)
	// N= # of total pedigrees
	// in the END: T = sum_N(D) / sqrt{sum_N(D^2)}  D of all N indeppendant pedigrees.
// only used in verbose mode
	int **sibs_aff;							// count for alleles in aff sibs
	int **sibs_una;							// count for alleles in un-aff sibs
	int **triads_trans;						// count for transmitted alleles in triads 
	int **triads_nottr;						// count for non-transmitted alleles in triads
	int **homo_par;							// count homozygot parents for each allele
};											// The storage requirements might change with the application...right now we need 2 floats.

class Consume_data{
public:
	Consume_data();
	~Consume_data();
	void init(int size, int id, struct Freq *, struct Freq*, int, int, int, int, int, int, int, int);
	void run();												// manages the synchronization
// the "_c" is for the check if typed version. It is slower. If you have perfect/simulated data use the above version
// If you have data where some markers are not typed use the "_c" version.
	void calculate_c(struct Fam&);							// do whatever calculations on the ith cell, add data to storage
	void check_ped(struct Fam&);							// will check pedigree for genotypic inconsistencies. Will print "x" for bad genotypes. 
															// Will "axe" entire nuclear family if one inconsistency is found
	int check_if_typed_nucfam(struct Fam& fa, int []);		// returns 1 if nuc. fam, 0 otherwise
	void calc_triads_c(int list[], struct Fam&);
	void calc_sibs_c(int list[], struct Fam&);
	void calc_geno_triads_c(int list[], struct Fam&);		// same as above but for the "geno" flavor of PDT
	void calc_geno_sibs_c(int list[], struct Fam&);
	int get_genotype(unsigned char, unsigned char, int);	// returns genotype given the 2 alleles, and the marker ID
	void haplotize(unsigned char* intype, unsigned char** outtype); // makes haplotype of size "window" out of the read in SNP data
	void explode_geno(struct Fam& fa);						// ONLY used if with check infile for genotypic inconsistencies.
															// explodes genotypes for entire pedigree into **mrkgeno
	int correct_nuc_fam(int []);							// checks nuclear family for genotyping errors. returns number of errors
															// The genotypes are stored in "**mrkgeno". Bad data will be marked as allele "254"
//int dbg[6];
//int dbgs[6];

private:
	int id;
	int store_d;											// 0|1 if to store d on a per SNP & pedigree basis
	int pedid;												// ID of current pedigree needed for data storage purposes. ID numbers are: 0,1,2,...
	int pdtave;												// usr PDT average stat 0|1
	int size;												// # of loci
	int verbose;											// collects a bunch of other info if this is set to 1
	int window;												// sliding window size for PDT, use 1 for regular PDT
	int ana_opt;											// analysis option
	int fam_count;
	int genopdt;											// 0== regular PDT, 1 == genopdt
	int check;												// 0 = assume each individual has either AL loci typed or no loci typed.
															// 1 = check for each locus if all memebers in dis.sib or triad are typed
	int *genostore;											// used in geno-PDT to temporarily store allele data
	struct Temp_storage storage;							// for data
	const struct Freq *fr;									// frequency data. READ-ONLY
	const struct Freq *hfr;									// haplotype data
	unsigned char *sib_list;								// keeps track which sibs have been counted in each pedigree
	unsigned char **indiv1;									// they hold the haplotypes for either the dis. sibpair (indiv1 & 2) or triads parents(=1&2), offspring(=3)
	unsigned char **indiv2;									// indiv[0] holds haplos on "left" chromosome, indiv[1] holds haplos on "right" chromosome...left & right is relative of course.
	unsigned char **indiv3;									// indiv1 is female (i.e., whoever has fa.geno[i][4] == 2). indiv2 is male, (i.e., whoever has fa.geno[i][4] == 1). in indiv3 the left allele always comes from mother, the right one from father (that's how SIMLA puts out the data)
	FILE *fptrped;											// ONLY used if with check infile for genotypic inconsistencies
	unsigned char **mrkgeno;								// ONLY used if with check infile for genotypic inconsistencies
	int max_ped_size;										// ONLY used if with check infile for genotypic inconsistencies
};

#endif
