#include "hashtable.h"

const int primes = 100;

int prlst[primes] = {3,5,7,11,13,17,19,23,29,
				31,37,41,43,47,53,59,61,67,
				71,73,79,83,89,97,101,103,107,109,113,
				127,131,137,139,149,151,157,163,167,173,
				179,181,191,193,197,199,211,223,227,229,
				233,239,241,251,257,263,269,271,277,281,283,
				293,307,311,313,317,331,337,347,349,353,
				359,367,373,379,383,389,397,401,409,419,
				421,431,433,439,443,449,457,461,463,467,
				479,487,491,499,503,509,521,523,541,547};

//**************************************************

HT::HT()
{
	HT::table = NULL;
	HT::entries = 0;
	HT::tablesize = 0;
}

//**************************************************

HT::~HT()
{
	int i;
	if(table != NULL){
		for(i = 0; i < tablesize; i++){
			if(table[i].key != NULL){
				delete [] table[i].key;
			}
		}
		delete [] table;
	}
	HT::entries = 0;
	HT::tablesize = 0;
}

//**************************************************

int HT::numberofentries()
{
	return(entries);
}

//**************************************************

void HT::inittable(int est_size)
{
	int i;
	if(tablesize == 0){
		HT::tablesize = next_larger_prime(est_size);
		HT::table = new Hrecord[tablesize];
		for(i = 0; i < tablesize; i++){
			table[i].key = NULL;
			table[i].value = -1;
		}
	}
	else{
		fprintf(stderr,"HT::inittable. Can only initialize hash table once.\n");
	}
}

//**************************************************

int HT::next_larger_prime(int guess)
{
	int candidate = (int)(guess * tablecoeff);
	int stop;
	if(0 == (candidate % 2)) candidate++;
	int found = 0;
	int div;
	int rem;
	int index;
	while(found == 0){
		index = 0;
		div = prlst[index];
		stop = (int)(sqrt((double)candidate));
		while(div < stop){
			rem = candidate % div;
			if(0 == rem){
				break;
			}
			index++;
			if(index >= primes){
				div += 2;
			}
			else{
				div = prlst[index];
			}
		} // while(div < stop){
		if(div >= stop){
			break;
		}
		candidate += 2;
	}
	return(candidate);
}

//**************************************************

int HT::insertrecord(char *item, int value)
{
	int index = hfn(item);
	int size;
	if(table[index].value == -1){										// key not in hashtable, add it.
		table[index].value = value;
		entries++;
		size = strlen(item) + 1;
		table[index].key = new char[size];
		strcpy(table[index].key, item);
	}
	else{																// space occupied, see if key is at a later position
		while(table[index].value >= 0){						
			if(0 == strcmp(item, table[index].key)){						// key already in table
				break;
			}
			index++;
			if(index == tablesize){
				index = 0;
			}
		}
		if(-1 == table[index].value){								
			table[index].value = value;
			entries++;
			size = strlen(item) + 1;
			table[index].key = new char[size];
			strcpy(table[index].key, item);	
		}
	}

	return(index);
}

//**************************************************

int HT::queryrecord(char *item)
{
	int index = hfn(item);
	int value = -1;
	int size;
	
	while(table[index].value >= 0){						
		if(0 == strcmp(item, table[index].key)){						// key already in table
			value = table[index].value;
			break;
		}
		index++;
		if(index == tablesize){
			index = 0;
		}
	}

	return(value);
}

//**************************************************

int HT::hfn(char *name)
{
	int pos = 0;
	int index = 0;
	int hashnum = 1;
	char chr = name[index];
	while(chr != '\0'){
		hashnum *= prlst[chr-44];
		hashnum %= tablesize;
		index++;
		chr = name[index];
	}
	return(hashnum);
}

//**************************************************

//**************************************************

//**************************************************

//**************************************************

//**************************************************

//**************************************************

//**************************************************
