#ifndef __MAIN_TDT__
#define __MAIN_TDT__

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>


#include "read_sync.h"
#include "consumer_sync.h"
#include "read_data.h"
#include "consume_data.h"
#include "sort.h"
#include "constants.h"
#include "version.h"
#include "help.h"
#include "cdflib.h"
#include "data.h"
#include "read_param.h"
#include "random.h"
#include "pathway.h"
#include "max.h"
#include "sync_random.h"
#include "hashtable.h"
#include <iostream>
#include <fstream>
using namespace std;

const double epsilon = 0.00000001;

#ifdef PROFILE
#include "profile.h"
#endif

struct Sync_pathway{
	int next_to_process;								// the locking thread will process this marker next and up date this variable. 
														// It will stop when "max_markers == next_to_process"
	pthread_mutex_t sync_pathway_mutex;						// 1 producer of random number, multiple consumers
};

struct Help_random{
	Random random_instance;
	uint32 seed;
};


/*
Requirements for PDT to work at all:
We need one or more triads (2 typed parents and an affected offspring) or one or more discordant (aff/unaff sibs) sib-pairs.

Step 1:
Counting alleles in triads and discordant sib-pairs within a single pedigree.
We have 4 parental alleles, call them p1a1 (like in parent 1, allele 1), p1a1, p2a1 and p2a2. We have aa1 & aa2 (affected allele1 & 2) 
and ua1 & ua2 (unaffected allele 1 & 2).

Allele counting in triads
count[aa1] += 2
count[aa2] += 2
count[p1a1] -= 1
count[p1a2] -= 1
count[p2a1] -= 1
count[p2a2] -= 1
So in effect, the 2 transmitted alleles get incremented by one while the non-transmitted get decremented by one. A zero sum game. 
The above way of counting appears the computationally most efficient way to accomplish this.
Similarly in discordant sib-pairs:
count[aa1] += 1
count[aa2] += 1
count[ua1] -= 1
count[ua2] -= 1
Again a zero sum game. The allele counts in the affected get incremented while the ones in the unaffected sib get decremented.
While doing all this we keep track of the number of triads (= nt) and the number of discordant sib-pairs (= ns). So for each allele 
(usually 2, but works for any number of alleles) we have a count from each of the sib-pairs (= xs) and from each of the triads (= xt).
Now we calculate: D = [1/(nt+ns)]*{sum_nt(xt) + sum_ns(xs)}, the score for ONE pedigree and one allele.

Step2: 
The Z-scrore.
Assume we have a total of n pedigrees. The score for the i_th pedigree is D_i. Now we calculate the allele specific z-score as follows:
Z = sum_n(D_i) / sqrt(sum_n(D_i^2))

Step 3:
The p-value:
Assume the locus under consideration has k (>= 2) alleles. Then we obviously have k-1 degrees of freedom and we use a CDF-chisq 
function. Let X = sum_alleles(Z^2), then the upper limit of integration using a chisq function with (k-1) degrees of freedom is 
U = (k-1)X / k.

###
Parallel processing considerations:
PDT processes one pdeigree at a time. It does not care about any other paedigrees or allele frequencies.
This algorithm has each thread processing one pedigree over all markers. Once it is done, it acquires a mutex and updates
a global count. It also acquires the ID of the next pedigree to be processed. Then it releases the mutex and processes the next pedigree
over all markers, etc.
*/
#endif
