#include "pathway.h"

pathway_class::pathway_class()
{
	index = new int;
	*index = maxran;
}

//***************************************

pathway_class::~pathway_class()
{

}

//***************************************

void pathway_class::init(struct pathway_info *arg)
{
	pathway_class::id = arg->id;
	pathway_class::permute_no = arg->id;
	pathway_class::num_loci = arg->num_loci;
	pathway_class::num_genes = arg->num_genes;
	pathway_class::num_pathway = arg->num_pathway;
	pathway_class::num_pedigrees = arg->num_pedigrees;
	pathway_class::permutations = arg->permutations;
	pathway_class::verbose = arg->verbose;
}

//***************************************

void pathway_class::run(struct pathway_info *arg)
{
	struct Record *list;
	init(arg);
	pdt_stats = NULL;
	pdt_stats = new float[num_loci];
	if(pdt_stats == NULL){
		fprintf(stderr,"Thread %i. Memory allocation for pdt_stats failed.\n", id);
		exit(0);
	}
	gene_score = new float[num_genes];
	gene_score_id = new float[num_genes];
	ranking = new unsigned short[num_genes];
	list = new struct Record[num_pathway];
	int j;

	while(permute_no < (permutations + 1)){
//		cout<<"Thread "<<id<<" is handling "<<permute_no<<"\n";
		if(verbose){
			fprintf(stderr,"Thread %i working on permutation %i\n",id,permute_no);
		}
		if(permute_no==0) {
			calculate_pdt_stats(num_loci,num_pedigrees);
			calculate_gene_scores(-1);
			calculate_ES_scores(-1);
		} else {
			permute(permute_no,num_pedigrees,num_loci);
			calculate_gene_scores(permute_no);
			calculate_ES_scores(permute_no);
		} // else


		pthread_mutex_lock(&sync_pathway->sync_pathway_mutex);	// lock mutex
		permute_no = sync_pathway->next_to_process;			// get next one to process
		sync_pathway->next_to_process++;					// update for next thread
		pthread_mutex_unlock(&sync_pathway->sync_pathway_mutex);
//fprintf(stderr,"endwhile(permute_no < (PER_LEN+1))\n");
	}


	delete [] pdt_stats;
	delete [] gene_score;
	delete [] gene_score_id;
	delete [] ranking;
	delete [] list;
}

//*******************************************

void pathway_class::calculate_pdt_stats(int num_loci,int num_fams) {
	int i,j;
	float temp;
	float numerator,denominator;
	for(i=0;i<num_loci;i++) {
		numerator = denominator = 0;
		for(j=0;j<num_fams;j++) {
		//for(j=0;j<50;j++) {
			temp = d_matrix[j][i];
			numerator += temp;
			denominator += temp*temp;
		}
		if(denominator<2)
			pdt_stats[i] = 0;
		else
			pdt_stats[i] = (numerator*numerator)/denominator;

	}
}

//*******************************************

void pathway_class::calculate_gene_scores(int permute) {
	int i,j,index,max_index;
	double max;
	int code,status;
	double p,q,df,bound,global;

	struct Record *list;
	list = new struct Record[num_genes];
	df = 1;
	FILE *fp;

	if(permute==-1) {
		fp = fopen("sig_snps_in_genes.txt","w");
	}
	for(i=0;i<num_genes;i++) {
		max = -1;
		j=0;
		while(g_index[i][j]!=-1) {
			index = g_index[i][j];
			if(pdt_stats[index]>max) {
				max = pdt_stats[index];
				max_index = index;
			}
			j++;
		}

		if(permute==-1) {
			fprintf(fp,"%s\n",rs_marker_names[max_index]);
		}

		pthread_mutex_lock(&sync_pathway->sync_pathway_mutex);
		code = 1;
		global = max;
		cdfchi(&code, &p, &q, &global, &df, &status, &bound);

		q = q*s_index[i];
		if(q>1)
			q = 1;

		p = 1-q;
		code = 2;
		cdfchi(&code, &p, &q, &global, &df, &status, &bound);
		
		list[i].id = i;
		list[i].cov = global;

		pthread_mutex_unlock(&sync_pathway->sync_pathway_mutex);
	}

	if(permute==-1) {
		fclose(fp);
	}

	Sort sorter;
	sorter.init(num_genes, list);
	sorter.qsort();

	j=0;
	for(i=num_genes-1;i>=0;i--) {
		//cout<<list[i].id<<" "<<list[i].cov<<"\n";
		gene_score[j] = list[i].cov;
		gene_score_id[j] = list[i].id;
		ranking[list[i].id] = j;
		j++;
	}

	delete [] list;
}

//*******************************************

void pathway_class::calculate_ES_scores(int permute) {
	int i,j,k,rank,index,Nh,max_index, geneindex;
	float stat,stat1,stat2,max = -1;
	float Nr,count, count1;
	float tmp;
	int *S = new int[num_genes];

	FILE *fp;
	
	if(permute==-1)
		fp = fopen("sig_genes.txt","w");

	for(i = 0; i < num_pathway; i++) {
		max = -9999;
//fprintf(stderr,"pathway_class::calculate_ES_scores. pathway %i of %i \n",i,num_pathway);
//#ifdef PROFILE
//    PROFILE_START("memset");
//#endif
//		memset((void*)S, -1, num_genes*sizeof(int));
//#ifdef PROFILE
 //   PROFILE_STOP("memset");
//#endif
		
		for(j = 0; j < num_genes; j++) {
/*
			// start old
			k = 0;
			count = 0;
			Nr = 0;
			while(p_index[i][k]!=-1) {   // k-th gene in the i-th pathway
				index = p_index[i][k];
				S[index] = 1;
				rank = ranking[index];
				if(rank<=j) {
					count=count + gene_score[rank];	
				}
				Nr = Nr + gene_score[rank];
				k++;
			}

			if(Nr!=0)
				stat1 = count/Nr;
			else
				stat1 = 0;

			Nh = k;

			count = 0;
			for(k=0;k<num_genes;k++) {
				if(S[k]!=1) {
					rank = ranking[k];
					if(rank<=j) {
						count++;	
					}
				}
			}

			if((num_genes-Nh)!=0) 
				stat2 = count/(double)(num_genes-Nh);
			else 
				stat2 = 0;

			stat = stat1-stat2;
			if(stat>max) {
				max = stat;
				max_index = j;
			}
		}
		if(permute==-1)
			ES[i] = max;
		else 
			ES_P[i][permute-1] = max;
// stop old
			*/
// start new
			k = 0;
			count1 = 0;
			Nr = 0;
			while(p_index[i][k] != -1) {   // k-th gene in the i-th pathway
				index = p_index[i][k];
				rank = ranking[index];
				tmp = gene_score[rank];
				if(rank<=j) {
					count1 += tmp;
				}
				Nr += tmp;
				k++;
			} // while(p_index[i][k]!=-1)


			if(Nr!=0)
				stat1 = count1 / Nr;
			else
				stat1 = 0;

			Nh = k;

			k = 0;
			count = j+1;
			while(index = p_index[i][k] != -1){
				index = p_index[i][k];
				rank = ranking[index];
				if(j>=rank) {
					count--;	
				}
				k++;
			} // for(k=0;k<num_genes;k++)

			if((num_genes-Nh)!=0) 
				stat2 = count/(float)(num_genes-Nh);
			else 
				stat2 = 0;

			stat = stat1-stat2;
			if(stat>max) {
				max = stat;
				max_index = j;
			}

		} // for(j=0;j<num_genes;j++) 
		if(permute==-1)
			ES[i] = max;
		else 
			ES_P[i][permute-1] = max;
// end new


		//// report sig genes in S
		if(permute==-1) {
			memset((void*)S, 1, num_genes*sizeof(int));
			k=0;
			fprintf(fp,"%s ", pathway_names[i]);
			while(p_index[i][k]!=-1) {   // k-th gene in the i-th pathway
				index = p_index[i][k];
				S[index] = 0;
				rank = ranking[index];
				if(rank<=max_index) {
				//	cout<<index<<" "<<gene_score[rank]<<"\n";
					fprintf(fp,"%d %f ",index,gene_score[rank]);
				}
				//cout<<"\n";
				k++;
			}
			fprintf(fp,"\n");
		} // if(permute==-1)
	} // for(i=0;i<num_pathway;i++)

	if(permute==-1)
		fclose(fp);
	delete [] S;
}

//*******************************************

void pathway_class::permute(int permute_no,int peds,int snps) {
	int j,k;
	float *sum,*denominator;
	float temp;
	sum = new float[snps];
	denominator = new float[snps];
	//double *randomnum = new double[peds];
	//Random ran;
	//ran.init();
	double random_num;

	//memset((void*)sum, 0, snps*sizeof(float));
	//memset((void*)denominator, 0, snps*sizeof(float));
	for(k = 0; k < snps; k++){
		sum[k] = 0;
		denominator[k] = 0;
	}	

	for(k = 0; k < peds; k++){
		//ran.bulk_ran1(peds,randomnum);
		//sum = 0;
		//denominator = 0;
		random_num = getran();
		//random_num = random_sets[permute_no-1][k];

		for(j = 0; j < snps; j++){
			temp = d_matrix[k][j];
			if(random_num > 0.5){
				sum[j] += temp;
			}
			else{
				sum[j] -= temp;
			}
			denominator[j] = denominator[j] + temp*temp;
		} // for(k = 0; k < peds; k++)
	}// for(j = 0; j < snps; j++)
	
	for(j = 0; j < snps; j++){
		if(denominator[j]>=2)
			pdt_stats[j] = (float)(sum[j]*sum[j])/denominator[j];
		else
			pdt_stats[j] = 0;
	}

	delete [] sum;
	delete [] denominator;
}

//*******************************************

double pathway_class::getran()
// responsible to get warehouse random numbers.
// we them from the random number thread (rannum_t) with associated mutex "sync_ran->sync_random_mutex"
// they come in "maxran" batches from :
/*
struct Sync_ran{
	int stop;									// if this is set to 1, thread should exit
	int ready;									// valid/ready entries 
	int empty;									// free entries...to be filled with random numbers
												// Note: maxranarrays - ready - empty = "cells" in use
	int list[maxranarrays];						// 0 = empty, 1 = ready, 2 = currently in use
	double ran_data[maxranarrays][maxran];		// where we store the data, see max.h [maxranarrays][maxran]
	pthread_mutex_t sync_random_mutex;			// 1 producer of random number, multiple consumers
	pthread_cond_t cond_waiting_producer;		// if random producer filled everthing up and is waiting on consumers to consume
	pthread_cond_t cond_waiting_consumer;		// if consumer is waiting on random producer to fill in some data
};
*/
{
	static int last_id = -1;					// ID of last batch of random number we got 0..(maxran-1)
	int i, j;
	static int temp = -1;
	if(*index == maxran){					// we ran out of inventory. let's get some more				
		i = 0;
		pthread_mutex_lock(&sync_ran->sync_random_mutex);				// acquire mutex
		while(sync_ran->ready == 0){									// Is data ready?
			pthread_cond_wait(&sync_ran->cond_waiting_consumer, &sync_ran->sync_random_mutex);		// If no data available wait for it
		}
		if(last_id >= 0){												// we need to return the ID of the previous batch to the pool of free cells
			sync_ran->list[last_id] = 0;								// return "last_id" cell to free pool
			temp = sync_ran->empty;										// so we might have to broadcast in case producer was sleeping due to "no empty cells"
			sync_ran->empty++;											// increment empty cell count
			j = 0;
		}
		while((sync_ran->list[i] != 1) && (i < maxranarrays)){			// find a cell that is ready
			i++;
		}
		if(i >= maxranarrays){											// this should never happen
			fprintf(stderr,"APL_class::getran. Could not get array of random numbers. Bailing out.\n");
			exit(0);
		}
		randomnumbers = sync_ran->ran_data[i];						
		sync_ran->list[i] = 2;
		sync_ran->ready--;
		if(temp == 0){
			pthread_cond_broadcast(&sync_ran->cond_waiting_producer);
		}
		last_id = i;													// remember which cell we're using so we can free it up next time around
		*index = 0;														// reset inventory count
		pthread_mutex_unlock(&sync_ran->sync_random_mutex);
	}
	return(randomnumbers[(*index)++]);
}

//*******************************************

//*******************************************
