#include "pedfiles.h"

//***********************************

PedFiles::PedFiles()
{
	infilelist = new char[maxnamesizepf];
	infileped = new char[maxnamesizepf];
	outfilelist= new char[maxnamesizepf];
	outfileped = new char[maxnamesizepf];
	bimfile = new char[maxnamesizepf];
	bedfile = new char[maxnamesizepf];
	famfile = new char[maxnamesizepf];
	non_marker_fields = -1;
	rows = -1;
	columns = -1;
	PedFiles::asciidata = NULL;
	PedFiles::maxlinesize = 0;
	PedFiles::line = NULL;
	PedFiles::one = NULL;
	PedFiles::two = NULL;
}

//***********************************

PedFiles::~PedFiles()
{
	int i;
	delete [] infilelist;
	delete [] infileped;
	delete [] outfilelist;
	delete [] outfileped;
	delete [] bimfile;
	delete [] bedfile;
	delete [] famfile;
	/*
	if(options & 1){
		for(i = 0; i < rows; i++){
			delete [] asciidata[i];
		}
		delete [] asciidata;
	}
	*/
	if(line != NULL){
		delete [] line;
	}
	if(one != NULL){
		delete [] one;
	}
	if(two != NULL){
		delete [] two;
	}
}

//***********************************

void PedFiles::run(struct Pedfinfo *arg)
{
	int i;
	strcpy(PedFiles::infileped, arg->infile_name_ped);
	strcpy(PedFiles::infilelist, arg->infile_name_list);
	strcpy(PedFiles::outfileped, arg->outfile_name_ped);
	strcpy(PedFiles::outfilelist, arg->outfile_name_list);
	strcpy(PedFiles::bedfile, arg->bedfile);
	strcpy(PedFiles::bimfile, arg->bimfile);
	strcpy(PedFiles::famfile, arg->famfile);
	PedFiles::options = arg->options;
	PedFiles::non_marker_fields = arg->non_marker_fields;


	if(options & 1){										// read .ped file and change to 0,1,2 genotype plus lookup table for conversion from 0,1,2 back to original
		getdimensions(columns, rows);			// get rows and colmns
		if((rows == 0) || (columns == 0)){
			fprintf(stderr,"PedFiles:: rows= %i columns= %i. Bailing out.\n", rows, columns);
			exit(0);
		}
		columns -= non_marker_fields;
		snps = columns / 2;
		one = new char[snps];
		two = new char[snps];
		memset((void*)one, ')',  snps*sizeof(char));
		memset((void*)two, ')',  snps*sizeof(char));
/*
		asciidata = new char*[rows];
		for(i = 0; i < rows; i++){
			asciidata[i] = new char[columns];
			memset((void*)asciidata[i], '0', columns*sizeof(char));
			
		}
*/
		readwritepedfile();									// read that mess from ASCII file

	}
	if(options & 2){										// read PLINK bim, bed & fam file and make standard .ped file for PDT, CAPL, etc.
		readwriteplink();
	}
}

//***********************************

void PedFiles::getdimensions(int &columns, int &rows)
{
	columns = 0;
	rows = 0;
	ifstream infile;
	string str;
	int linesize;
	char c;
	char *token;
	
	int charactersperline = 0;

	infile.open(PedFiles::infileped, ios::in);
	if(infile.is_open()){
		infile.clear();
		infile.seekg(0, ios::beg);
		while((infile.get(c)) && (!iscntrl((int)c))){
			charactersperline++;
		}
		if(charactersperline > 0){
			infile.clear();
			infile.seekg(0, ios::beg);
			maxlinesize = 4*charactersperline;
			if(line == NULL){
				line = new char[maxlinesize];
			}
			if(line == NULL){
				fprintf(stderr,"PedFiles::getdimensions(), memory allocation error.\n");
				exit(0);			
			}
			infile.getline(line, maxlinesize);				// now read one line to get the number of fields
			token = strtok(line, " ");
			columns = 1;
			while(NULL != strtok(NULL, " ")){
				columns++;
			}
			rows = 1;
			while(infile.getline(line, maxlinesize)){
				rows++;
				fprintf(stderr,"%i\n",rows);
			}
		}
	}
	else{
		fprintf(stderr,"PedFiles::getdimensions(), Can not open %s. Bailing out.\n", infileped);
		exit(0);
	}
	infile.close();
}

//***********************************
/*
     0000000: 01101100 00011011 00000001 11011100 00001111 11100111  l.....
     0000006: 00001111 01101011 00000001                             .k.
The actual binary data are the nine blocks of 8 bits (a byte) in the center: the first 3 bytes have a special meaning. The first two are fixed, a 'magic number' that enables PLINK to confirm that a BED file is really a BED file. That is, BED files should always start 01101100 00011011. The third byte indicates whether the BED file is in SNP-major or individual-major mode: a value of 00000001 indicates SNP-major (i.e. list all individuals for first SNP, all individuals for second SNP, etc) whereas a value of 00000000 indicates individual-major (i.e. list all SNPs for the first individual, list all SNPs for the second individual, etc). By default, all BED files are SNP-major mode (as is the example below). 
Here we have extracted and annotated the relevant part of the xxd output: 
     |-magic number--| |-mode-| |--genotype data---------| 

     01101100 00011011 00000001 11011100 00001111 11100111

     |--genotype data-cont'd--| 

     00001111 01101011 00000001 

For the genotype data, each byte encodes up to four genotypes (2 bits per genoytpe). The coding is 
     00  Homozygote "1"/"1"
     01  Heterozygote
     11  Homozygote "2"/"2"
     10  Missing genotype
The only slightly confusing wrinkle is that each byte is effectively read backwards. That is, if we label each of the 8 position as A to H, we would label backwards: 
     01101100
     HGFEDCBA
and so the first four genotypes are read as follows: 
     01101100
     HGFEDCBA

           AB   00  -- homozygote (first)
         CD     11  -- other homozygote (second)
       EF       01  -- heterozygote (third)
     GH         10  -- missing genotype (fourth)
Finally, when we reach the end of a SNP (or if in individual-mode, the end of an individual) we skip to the start of a new byte (i.e. skip any remaining bits in that byte). 
It is important to remember that the files test.bim and test.fam will already have been read in, so PLINK knows how many SNPs and individuals to expect. 
So, considering the full test.bed file: here we consider the six bytes that contain all the genotype data. We consider them one at a time, showing how the 4 genotypes are extracted from each byte to make up the entire dataset. Some positions are called null meaning that all the genotypes for that SNP have been read in, so we advance to the start of a new byte for the next SNP (when in SNP-major mode): 

                Genotype    Person    SNP
     11011100 

           00   G/G         1 1       snp1
         11     A/A         1 2       snp1
       10       0/0         1 3       snp1
     11         A/A         2 1       snp1


     00001111 

           11   A/A         2 2       snp1
         11     A/A         2 3       snp1
       00       (null)
     00         (null)


     11100111
           
           11   2/2         1 1       snp2
         10     0/0         1 2       snp2
       01       1/2         1 3       snp2
     11         2/2         2 1       snp2


     00001111 
  
           11   2/2         2 2       snp2
         11     2/2         2 3       snp2
       00       (null) 
     00         (null)


     01101011

           11   C/C         1 1       snp3
         01     A/C         1 2       snp3
       01       A/C         1 3       snp3
     10         0/0         2 1       snp3


     00000001

           10   0/0         2 2       snp3
         00     A/A         2 3       snp3
       00       (null)
     00         (null)

In summary, the following define the BED file format 
	First two bytes 01101100 00011011 for PLINK v1.00 BED file 
	Third byte is 00000001 (SNP-major) or 00000000 (individual-major) 
	Genotype data, either in SNP-major or individual-major order 
	New "row" always starts a new byte 
	Each byte encodes up to 4 genotypes 
	10 indicates missing genotype, otherwise 0 and 1 point to allele 1 or allele 2 in the BIM file, respectively 
	Bits in each byte read in reverse order 
Any changes to this format will be accompanied by a different, unique magic number and will be backwards compatabile in PLINK 
Old versions Earlier versions: v0.99 BED files do not contain the 2-byte magic number; BED files prior to 0.99 are always in individual-major mode and contain neither the magic-numbers nor the SNP-major/individual-major identifier. PLINK will indicate if these earlier, legacy files are found.



*/
void PedFiles::readwriteplink()
{
	int i,j,k,m;
	int allele1, allele2;
	ifstream infilefam;
	ifstream infilebin;
	ofstream outfile;
	char *newline = new char[1000];
	char *token;	
	unsigned char data;
	unsigned char geno[4];

	PedFiles::snps = getsnps();										// number of SNPs
	PedFiles::indiv = getindiv();									// number of individuals
	one = new char[snps];
	two = new char[snps];
	int offset = indiv / 4;
	if((offset*4) < indiv){
		offset++;
	}
	char **lines = new char*[4];
	int linesize = (4*snps)+2;
	for(i = 0; i < 4; i++){
		lines[i] = new char[linesize];
	}
	int position = 0;
	infilebin.open(PedFiles::bedfile, ios::in | ios::binary);										
	infilebin.seekg(0, ios::end);
	int filesize = infilebin.tellg();
	filesize -= plinkoffset;
	infilebin.seekg(plinkoffset, ios::beg);
	infilebin.clear();
	char *buffer = new char[filesize];								// we will read 1 KB from the binary file at a time, then process it, then read another KB, etc 
	infilebin.read(buffer, filesize);
	infilebin.close();


	infilefam.open(PedFiles::famfile, ios::in);
	infilefam.clear();										
	infilefam.seekg(0, ios::beg);									// it should do "clear()" and "seekg(0, ios::beg)" as default when opening a file, but I have seen instances where it didn't do that
	outfile.open(PedFiles::outfileped);
	for(i = 0, k = 0; i < indiv; i += 4, k++){									// on byte contains info for 4 SNPs at a time. So, to be more efficient, we need to process 4 individuals at a time
		for(m = 0; m < 4; m++){
//			strcpy(lines[m], "");									// reset each genotype line to empty string
			memset((void*)lines[m], (int)' ', sizeof(char)*linesize);
			lines[m][4*snps] = '\0';
		}
		for(j = 0; j < snps; j++){
			data = buffer[k+(j*offset)];							// genotypes: 0=(1,1); 1=(1,2); 2=(0,0); 3=(2,2) 
			geno[0] = 0x03 & data;
			geno[1] = 0x0C & data;
			geno[1] >>= 2;
			geno[2] = 0x30 & data;
			geno[2] >>= 4;
			geno[3] = 0xC0 & data;
			geno[3] >>= 6;
			allele1 = 4*j + 1;
			allele2 = 4*j + 3;
			for(m = 0; m < 4; m++){
				/*
				switch(geno[m]){
					case 2:
						strcat(lines[m], " 1 2");
						break;
					case 0:
						strcat(lines[m], " 1 1");
						break;
					case 3:
						strcat(lines[m], " 2 2");
						break;
					case 1:
						strcat(lines[m], " 0 0");
						break;
					default:
						fprintf(stderr,"PedFiles::readwriteplink. Genotype of %i\n",geno[m]);
						break;
				}
				*/
				switch(geno[m]){
					case 2:
						lines[m][allele1] = '1';
						lines[m][allele2] = '2';
						break;
					case 0:
						lines[m][allele1] = '1';
						lines[m][allele2] = '1';
						break;
					case 3:
						lines[m][allele1] = '2';
						lines[m][allele2] = '2';
						break;
					case 1:
						lines[m][allele1] = '0';
						lines[m][allele2] = '0';
						break;
					default:
						fprintf(stderr,"PedFiles::readwriteplink. Genotype of %i\n",geno[m]);
						break;
				}				
			}
		}
		infilefam.getline(newline, 1000);
		outfile <<newline<<" "<<lines[0]<<endl;
		if(!infilefam.eof()){
			infilefam.getline(newline, 1000);
			outfile <<newline<<" "<<lines[1]<<endl;
		}
		if(!infilefam.eof()){
			infilefam.getline(newline, 1000);
			outfile <<newline<<" "<<lines[2]<<endl;
		}
		if(!infilefam.eof()){
			infilefam.getline(newline, 1000);
			outfile <<newline<<" "<<lines[3]<<endl;
		}
	}
	outfile.close();
	infilefam.close();
	delete [] buffer;
	delete [] newline;
	for(i = 0; i < 4; i++){
		delete [] lines[i];
	}
	delete [] lines;
}

//***********************************

int PedFiles::getsnps()
{
	int result = 0;
	ifstream infile;
	infile.open(PedFiles::bimfile, ios::in);
	infile.clear();										
	infile.seekg(0, ios::beg);
	char *temp = new char[1024];							// just a dummy to read lines
	while(infile.getline(temp, 1023)){
		result++;
	}
	infile.close();
	delete [] temp;
	return(result);
}

//***********************************

int PedFiles::getindiv()
{
	int result = 0;
	ifstream infile;
	infile.open(PedFiles::famfile, ios::in);
	infile.clear();										
	infile.seekg(0, ios::beg);
	char *temp = new char[1024];							// just a dummy to read lines
	while(infile.getline(temp, 1023)){
		result++;
	}
	infile.close();
	delete [] temp;
	return(result);
}

//***********************************

void PedFiles::readwritepedfile()
{
	int i;
	ifstream infile;
	ofstream outfile;
	char *newline = NULL;
	char *token;
	char allele;
	int allele_count;
	int snp;
	int index;
	int linenum = 0;
	unsigned int previous;
	newline = new char[2*columns + 20*non_marker_fields + 100];
	outfile.open(PedFiles::outfileped);
	infile.open(PedFiles::infileped, ios::in);
	if(infile.is_open()){
		infile.clear();
		infile.seekg(0, ios::beg);
		while(infile.getline(line, maxlinesize)){
//fprintf(stderr,"%i\n",linenum);
/*
this works well but is way too slow
			strcpy(newline, "\0");
			token = strtok(line, " ");
			strcat(newline, token);
			strcat(newline, " ");
			for(i = 1; i < non_marker_fields; i++){
				token = strtok(NULL, " ");
				strcat(newline, token);
				strcat(newline, " ");
			}
			allele_count = 0;
			while(token = strtok(NULL, " ")){
				allele = token[0];
				snp = allele_count / 2;
				if(!iscntrl((int)allele)){			// at this point the token represents some valid entry, not "\n"
					if(allele == one[snp]){
						strcat(newline, " 1");
					}
					else if(allele == two[snp]){
						strcat(newline, " 2");
					}
					else if(allele == '0'){
						strcat(newline, " 0");
					}
					else if(one[snp] == ')'){
						one[snp] = allele;
						strcat(newline, " 1");
					}
					else if(two[snp] == ')'){
						two[snp] = allele;
						strcat(newline, " 2");
					}
					else{
						fprintf(stderr,"PedFiles::readwritepedfile. Unexpected character at line %i and column %i\n", linenum+1, non_marker_fields+allele_count+1);
						strcat(newline, " 0");
					}
				}
				allele_count++;
			} // while(token = strtok(NULL, " "))
	*/
			allele_count = -1 * non_marker_fields;
			index = 0;
			previous = 1;
			while(allele_count < 0){				// eat through the header to get to alleles
				if(line[index] != ' '){
					previous = 1;
				}
				else if(previous == 1){
					previous = 0;
					allele_count++;
				}
				else{
					previous = 0;
				}
//cout <<line[index];
				index++;
			}
			while(line[index] != '\0'){
				allele = line[index];
				if(allele != ' '){
					if(allele != '0'){
						snp = allele_count / 2;
						if(allele == one[snp]){
							line[index] = 49;
						}
						else if(allele == two[snp]){
							line[index] = 50;
						}
						else if(one[snp] == ')'){
							one[snp] = allele;
							line[index] = 49;
						}
						else if(two[snp] == ')'){
							two[snp] = allele;
							line[index] = 50;
						}
						else{
							fprintf(stderr,"PedFiles::readwritepedfile. Unexpected character at line %i and column %i\n", linenum+1, index+1);
						}
					}
					allele_count++;
				}
				index++;
			}
//			outfile <<newline<<endl;
			outfile <<line<<endl;
			linenum++;
		} // while(infile.getline(line, maxlinesize)){
	}
	else{
		fprintf(stderr,"PedFiles::readpedfile(), Can not open %s. Bailing out.\n", infileped);
		exit(0);
	}
	infile.close();
	outfile.close();
	delete [] newline;
	outfile.open(PedFiles::outfilelist);
	outfile <<snp+1<<endl;
	for(i = 0; i < (snp+1); i++){
		outfile <<one[i]<<" "<<two[i]<<endl;
	}
	outfile.close();
}

//***********************************

void PedFiles::bim2map(char *outfilename)
{
	int i = 1;
	char *temp = new char[501];
	char *token;
	ifstream infile;
	ofstream outfile;
	outfile.open(outfilename);
	infile.open(PedFiles::bimfile, ios::in);
	if(infile.is_open()){
		infile.clear();
		infile.seekg(0, ios::beg);
		while(infile.getline(temp, 500)){
			token = strtok(temp, "\t");
			token = strtok(NULL, "\t");
			outfile <<"1 "<<token<<" "<<i<<endl;
			i++;
		}
	}
	else{
		fprintf(stderr,"PedFiles::bim2map(), Can not open %s. Bailing out.\n", infileped);
		exit(0);
	}
	infile.close();
	outfile.close();
	delete [] temp;
}

//***********************************
