#include "read_data.h"

//*********************************

Read_data::Read_data()
{
}

//*********************************

Read_data::~Read_data()
{
	if(line){
		delete [] line;
	}
}

//*********************************

void Read_data::init(char *name, int l, int s, int n, int mf)
{
	strcpy(infile, name);
	Read_data::loci = l;
	Read_data::size = s + freebe;
	Read_data::non_allele_fields = n;
	Read_data::max_fam_size = mf;
	line = new char[size];
//fprintf(stderr,"reader::init: size of line = %i\n",size);
	strcpy(line,"-1");						// so we know that it doesn't hold any valid data yet
}

//*********************************
// the synchronization structure below
//struct Read_sync{							// structure that tells use where the data is, what is used, what is ready and what is empty
//	int done;								// initialized to 0, when EOF is encountered the reader sets it to 1 and returns a "not genotyped" Fam upon call to next().
//											// Consumers decremts "thread_count" and exits. Upon a "thread_count" of 0 the reader exits as well.
//	int thread_count;						// Number of active consumer threads. Return when this number reaches 0.
//	int size;								// how many items (i.e. struct Fam) are in list
//	int free;								// how many items are ready to be filled by producer (i.e. reader)
//	int ready;								// how many items are filled by producer, waiting to be consumed
//	int *list;								// status of each "struct Fam".		-1 = free, ready to recieve data
//											//									0  = ready, filled with data, waiting to be consumed
//											//									1  = currently being used by consumer or producer
//	pthread_mutex_t mut_reader;				// lock if "Read_sync" is updated, that's between ONE reader and 1+ consumer(s)
//	pthread_cond_t cond_waiting_producer;	// if reader is waiting on consumers
//	pthread_cond_t cond_waiting_consumer;	// if consumer is waiting on producer/reader
//};
void Read_data::run()
{
	int i;
	int temp;
	int lines = -1;
	ifstream inf;
	inf.open(infile, ios::in);
	if(!inf.is_open()){
		fprintf(stderr,"Read_data::run. Can not open %s, bailing out.\n",infile);
		exit(0);
	}
	// FILE *fptr = open(infile,"r");								// somehow this doesn't work with the Microsoft compiler VS2005
	while(read_synchronizer->done == 0){							// the read will set this to "1" when EOF is encountered, at which point all data has been read and the reader is done.
																	// only this single reader manipulates "done", so no need to lock mutex
		pthread_mutex_lock(&read_synchronizer->mut_reader);
		while(read_synchronizer->free == 0){						// there's no place to put anymore data. We need to wait for the consumers to use up some of the data
			pthread_cond_wait(&read_synchronizer->cond_waiting_producer, &read_synchronizer->mut_reader);
		}
		i = 0;
		while((i < read_synchronizer->size) && (read_synchronizer->list[i] != -1)){
			i++;
		}
		if(i == read_synchronizer->size){
			fprintf(stderr,"Read_data::run(): no free Fam, error in  read_synchronizer. Bailing out.\n");
			exit(0);
		}
		read_synchronizer->list[i] = 1;								// this indicates we are using the ith slot
		read_synchronizer->free--;									// decrement the number of free cells
		pthread_mutex_unlock(&read_synchronizer->mut_reader);		// release the lock, since we updated the sync. structure
		// now do some reading...after all, that's what we're here for
//		lines = read_fam(i,fptr);										// now read data from infile into the ith cell
		lines = read_fam(i, inf);
		pthread_mutex_lock(&read_synchronizer->mut_reader);			// re-acquire the lock to update the sync structure again.
		if(lines > 0){
			// update sync structure
			temp = read_synchronizer->ready;
			read_synchronizer->list[i] = 0;							// indicate that the ith cell is ready to be consumed, it has fresh, valid data
			read_synchronizer->ready++;								// ...opposite of "free"
			if(temp == 0){											// we better wake up sleeping consumers waiting on data
				pthread_cond_broadcast(&read_synchronizer->cond_waiting_consumer);
			}
		}
		else{
			read_synchronizer->done = 1;							// Can't read no more, either error or EOF
		}
		pthread_mutex_unlock(&read_synchronizer->mut_reader);
	}
	pthread_cond_broadcast(&read_synchronizer->cond_waiting_consumer);
//	fprintf(stderr,"Read_data::run(): done reading\n");
	//fclose(fptr);
	inf.close();
}

//*********************************

int Read_data::read_fam(int i, /*FILE *fptr*/ifstream& inf)
//private:
//	FILE *fptr;
//	int loci;				// number of loci
//	char *line;				// this initially holds the line read in from the .ped file
//							// since alleles are sperated by ONE (not 2 or 3 or more) whitespace
//							// will will have 4 bytes/char per locus, plus pedigree ID and all the other stuff
//							// at the beginning of each line...say 100 char, max. So total we would have
//							// "freebe + 4 * loci" total characters in a line, where loci might be up to 50000.
//	int size;				// = freebe + 4 * loci
{
#if 0						// works, but slow
	static int famid = -1;
	int lines_read = 0;
	int j;
	int k,l;
	int currid;
	unsigned char a1;											// allele1
	unsigned char a2;											// allele2
	unsigned int st;
	if(famid == -1){
		inf >> famid;
	}
	currid = famid;
	families[i].members = 0;
	while(currid == famid){
		if(inf.eof()){
			read_synchronizer->done = 1;						// Can't read no more, either error or EOF
			return(lines_read);									// the last line read contains nothing, might as well return
		}
		for(j = 1; j < 6; j++){
			inf >> st;
			families[i].geno[lines_read][j] = (unsigned char)st;
		}
		for(j = 0; j < families[i].loci; j++){
			inf >> st;
			a1 = (unsigned char)st;
			inf >> st;
			a2 = (unsigned char)st;
			a1 <<= 4;												// left shift 4 bits
			families[i].geno[lines_read][6+j] = a1 + a2;									// we store allele1 in upper 4 bits, allele2 in lower 4bits
		}
		lines_read++;
		families[i].members++;
		inf >> famid;											// last thing is to read FamiID to see if we're still within same pedigree. If yes, caryy on, if no return
	}
	return(lines_read);
#endif
#if 1										// much faster than above approech
	
	int famid = -1;												// we're not sure how many lines the same pedigree spans
	int currid = -1;
	int lines_read = 0;
	famid = atoi(line);											// if(famid == -1) we are at the very beginning of infile,i.e. read first line. Otherwise use line already in buffer
	families[i].members = 0;
	if(famid == -1){
		if(inf.getline(line,size)){
			famid = atoi(line);
		}
	}
	families[i].id = famid;
	currid = famid;		
	// now we keep on reading & processing until famid != previd
	while((currid == famid) && (currid > 0)){
		if(Read_data::max_fam_size == lines_read){
			fprintf(stderr,"Read_data::read_fam. Maximum family size of %i exceeded. Good bye.\n", max_fam_size);
			exit(0);
		}
		if(process(line, families[i].geno[lines_read])){
			lines_read++;
			families[i].members++;
			inf.getline(line,size);
			if(inf.eof()){
				read_synchronizer->done = 1;						// Can't read no more, either error or EOF
				return(lines_read);									// the last line read contains nothing, might as well return
			}
		}
		famid = atoi(line);										// as along as "famid == currid" we keep things going
// Can we use atoi() when only the first part of the line represents the famid? Yes, see below.
////// int atoi(const char *) ///////////
// Convert string to integer
// Parses the C string str interpreting its content as an integral number, which is returned as an int value.
// The function first discards as many whitespace characters as necessary until the first non-whitespace character is found. Then, starting from this character, takes an optional initial plus or minus sign followed by as many numerical digits as possible, and interprets them as a numerical value.
// The string can contain additional characters after those that form the integral number, which are ignored and have no effect on the behavior of this function.
/////////////////////////////////////////
	}
	return(lines_read);											// hopefully > 0, otherwise error or EOF
#endif

	}

//*********************************

int Read_data::process(char* raw, unsigned char* cooked)
//struct Fam{								// structure that holds the data
//	int members;						// each Family has 2 parents and "members - 2" children
//	int loci;							// number of loci, kind of redundant since all Fams are going to have the same number
//	unsigned char **geno;				// geno[i][j] is for the ith member and jth locus
//										// geno[i][0] holds 0|1. 0 = not genotyped, 1 = valid data
//										// geno[members][loci+1]
										// geno[members][loci+6]
										// geno[i][1] = "individual ID"
										// geno[i][2] = "1st parent"
										// geno[i][3] = "2nd parent"
										// geno[i][4] = "male=1,female=2"
										// geno[i][5] = "aff.stat. aff=1,unaff=2,unkown=0"
										// geno[i][6+] = "upper 4 bits allele1, lower 4 bits allele2"
//};

{
	unsigned char a1;											// allele1
	unsigned char a2;											// allele2
	unsigned int ui32;
	char *tokenptr;
	char *token1;												// read allele 1
	char *token2;												// read allele 2
	int i = 0;
	famid = strtok(raw, " ");									// famiID, not needed
//fprintf(stderr,"%s\n",famid);
	for(i = 1; i < non_allele_fields; i++){										// keep [0] open for now
		tokenptr = strtok(NULL, " ");
		ui32 = atoi(tokenptr);
		ui32 &= 0x000000ff;										// basically modulus 256
		cooked[i] = (unsigned char)ui32;
//		cooked[i] = (unsigned char)atoi(tokenptr);
	}
	i = 0;														// let's count the number of loci
	// in case "non_allele_fields > 6" there's garbage info we don't care about. 
	// It will be overwritten below. Fields [0] - []
	while(((token1 = strtok(NULL, " ")) != NULL) && ((token2 = strtok(NULL, " ")) != NULL)){
		a1 = (unsigned char)atoi(token1);
		a2 = (unsigned char)atoi(token2);
		a1 <<= 4;												// left shift 4 bits
		cooked[6+i] = a1 | a2;									// we store allele1 in upper 4 bits, allele2 in lower 4bits
		i++;
	}
	if(i != loci){
		fprintf(stderr,"Read_data::process(). Expected %i loci, read %i loci. Famid = %s\n", loci, i, famid);
	}
	return(i);
}

//*********************************
