#ifndef __READ_SYNC__
#define __READ_SYNC__

#ifdef WIN32
#include "pthread.h"
#include "sched.h"
#include "semaphore.h"
#include <windows.h>
#endif
#ifdef SOLARIS
#include <pthread.h>			// define _REENTRANT in makefile
#include <string.h>
#endif
#ifdef LINUX
#include <pthread.h>			// define _REENTRANT in makefile
#include <string.h>
#endif

struct Fam{								// structure that holds the data
	int id;								// family identifier
	int members;						// each Family has 2 parents and "members - 2" children
	int loci;							// number of loci, kind of redundant since all Fams are going to have the same number
	unsigned char **geno;				// geno[i][j] is for the ith member and jth locus
										// geno[i][0] holds 0|1. 0 = not genotyped, 1 = valid data
										// geno[members][loci+6]
										// geno[i][1] = "individual ID"
										// geno[i][2] = "1st parent"
										// geno[i][3] = "2nd parent"
										// geno[i][4] = "male=1,female=2"
										// geno[i][5] = "aff.stat. aff=1,unaff=2,unkown=0"
										// geno[i][6+] = "upper 4 bits allele1, lower 4 bits allele2"

};

struct Read_sync{							// structure that tells use where the data is, what is used, what is ready and what is empty
	int done;								// initialized to 0, when EOF is encountered the reader sets it to 1 and returns a "not genotyped" Fam upon call to next().
											// Consumers decremts "thread_count" and exits. Upon a "thread_count" of 0 the reader exits as well.
	int thread_count;						// Number of active consumer threads. Return when this number reaches 0.
	int size;								// how many items (i.e. struct Fam) are in list
	int free;								// how many items are ready to be filled by producer (i.e. reader)
	int ready;								// how many items are filled by producer, waiting to be consumed
	int ped_id;								// current pedigree Id, {0,1,2,...,max_ped-1}, max_ped = number of pedigrees in infile.ped
	int *list;								// status of each "struct Fam".		-1 = free, ready to recieve data
											//									0  = ready, filled with data, waiting to be consumed
											//									1  = currently being used by consumer
	pthread_mutex_t mut_reader;				// lock if "Read_sync" is updated, that's between ONE reader and 1+ consumer(s)
	pthread_cond_t cond_waiting_producer;	// if reader is waiting on consumers
	pthread_cond_t cond_waiting_consumer;	// if consumer is waiting on producer/reader
};

#endif
